#version 330 compatibility

#info See blog post: <https://mathr.co.uk/blog/2018-11-17_newtons_method_for_periodic_points.html>

#include "Progressive2D.frag"
#include "Complex.frag"

uniform vec2 Center;
uniform float Zoom;

#group Mandelbrot
// period of nuclei to find
uniform int Period; slider[1,1,64]
// number of iterations for Newton's method
uniform int NewtonSteps; slider[1,64,64]
// number of iterations for exterior DE colouring
uniform int Iterations; slider[1,64,4096]

const float phi = 1.618033988749895;

// http://lolengine.net/blog/2013/07/27/rgb-to-hsv-in-glsl
vec3 hsv2rgb(vec3 c)
{
  vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
  vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
  return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

// find periodic nucleus using Newton's method
vec2 Mandelbrot_nucleus(vec2 c0, int period, int steps, bool new)
{
  vec2 c = c0;
  for (int j = 0; j < steps; ++j)
  {
    vec4 G = cConst(1.0);
    vec4 z = vec4(0.0);
    for (int l = 1; l <= period; ++l)
    {
      z = cSqr(z) + cVar(c);
      if (l < period && period % l == 0 && new)
        G = cMul(z, G);
    }
    G = cDiv(z, G);
    c -= cDiv(G.xy, G.zw);
  }
  return c;
}

// find exterior distance estimate
float Mandelbrot_exterior_de(vec2 c0, int n)
{
  vec4 z = cConst(0.0);
  vec4 c = cVar(c0);
  for (int i = 0; i < n; ++i)
  {
    z = cSqr(z) + c;
    if (cNorm(z.xy) > 65536.0)
      break;
  }
  if (cNorm(z.xy) > 65536.0)
    return length(z.xy) * log(length(z.xy)) / length(z.zw);
  return -1.0;
}

// main entry point
vec3 color(vec2 z0)
{
  vec2 z = Mandelbrot_nucleus(z0, Period, NewtonSteps, z0.y < 0.0);
  float de = Mandelbrot_exterior_de(z0, Iterations);

  // hue pseudorandom from nucleus location
  vec2 z1 = cLog((z - Center) * Zoom);
  float hue = abs(z1.y) + phi * z1.x;
  hue -= floor(hue);

  // saturated near nucleus locations
  float d = distance(z, z0) * Zoom * float(Period);
  float sat = 1.0 / (1.0 + d);

  // white Newton basin boundaries
  float inD = length(vec4(dFdx(z0), dFdy(z0)));
  float outD = length(vec4(dFdx(z), dFdy(z)));
  if (! (outD / inD < 1.0)) sat = 0.0;

  // dark Mandelbrot set boundary
  float b = de / inD;
  float val = b < 0.0 ? 1.0 : tanh(clamp(b, 0.0, 4.0));

  // bad value checks
  if (!(z1 == z1)) return vec3(1.0);
  if (cNorm(z) > 4.0 || !(z == z)) return vec3(1.0);
  if (isinf(de) || isnan(de)) return vec3(1.0);
  if (isinf(b) || isnan(b)) return vec3(1.0);

  return hsv2rgb(vec3(hue, sat, val));
}

#preset Period2
Center = -0.75,0
Zoom = 0.7
Period = 2
#endpreset

#preset Period3
Center = -0.75,0
Zoom = 0.7
Period = 3
#endpreset

#preset Period4
Center = -0.75,0
Zoom = 0.7
Period = 4
#endpreset

#preset Period9
Center = -1.76,0
Zoom = 20
Period = 9
#endpreset

#preset Period12
Center = -1.76,0
Zoom = 20
Period = 12
#endpreset
